﻿#region Using directives

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Runtime.Serialization;
using System.Text;

#endregion

namespace Chapter_17
{
    #region Generic Type Constraint Demo

    // customers

    public class Customer : ISerializable
    {
        #region ISerializable Members

        public void GetObjectData(SerializationInfo info, StreamingContext context)
        {
            throw new NotImplementedException();
        }

        #endregion
    }

    public class LargBiz : Customer { }
    public class SmallBiz : Customer { }
    public class HomeUser : Customer { }

    // orders

    public class Order : ISerializable
    {
        #region ISerializable Members

        public void GetObjectData(SerializationInfo info, StreamingContext context)
        {
            throw new NotImplementedException();
        }

        #endregion
    }

    public class MailOrder : Order { }
    public class RetailOrder : Order { }

    /// <summary>
    /// Shows how to use constraints on a type.
    /// </summary>
    /// <typeparam name="V">Customer and ISerializable</typeparam>
    /// <typeparam name="W">Order and ISerializable</typeparam>
    public class CustomerProcessor<V, W>
        where V : Customer, ISerializable
        where W : Order, ISerializable
    {
        public void ProcessOrder(V customer, W order)
        {
            if (!ValidateAddress(customer)) return;
            if (!ValidateCredit(customer)) return;
            if (!ValidateAvailability(order)) return;

            PlaceOrder(customer, order);

            Console.WriteLine("Order Processed!");
        }

        protected virtual bool ValidateAddress(V customer)
        {
            // do something to validate an address
            return true;
        }

        protected virtual bool ValidateCredit(V customer)
        {
            // do something to validate customer credit
            return true;
        }


        protected virtual bool ValidateAvailability(W order)
        {
            // do something to validate order availability
            return true;
        }

        protected virtual void PlaceOrder(V customer, W order)
        {
            // do something to place the order
        }
    }

    #endregion

    /// <summary>
    /// Demo for working with various types of constraints
    /// </summary>
    public class ConstraintDemo
    {
        public void DoSomething<T>(T parm)
        {
        }

        /// <summary>
        /// Type that does not implement IEnumerable<T>.
        /// Used for TestInterfaceConstraints Demo.
        /// </summary>
        /// <typeparam name="T">Type to hold.</typeparam>
        class MyType<T> : IEnumerable
        {
            // internal data store
            private List<T> myList = new List<T>();

            // add item to list
            public void Add(T item)
            {
                myList.Add(item);
            }

            // access list via index
            public T this[int index]
            {
                get
                {
                    return myList[index];
                }
                set
                {
                    myList[index] = value;
                }
            }

            #region IEnumerable Members

            // wraps myList GetEnumerator
            public IEnumerator GetEnumerator()
            {
                return myList.GetEnumerator();
            }

            #endregion
        }

        /// <summary>
        /// Algorithm to demonstrate and test
        /// interface constraints.
        /// </summary>
        public void TestInterfaceConstraints()
        {
            // list of value types is okay
            List<int> myInts = new List<int>();
            myInts.AddRange(new int[] { 1, 2, 3, 4, 5, 6 });

            PrintEnumerableValues<List<int>, int>(myInts);

            // list of reference types is okay
            List<string> myStrings = new List<string>();
            myStrings.AddRange(new string[] { "one", "two", "three" });

            PrintEnumerableValues<List<string>, string>(myStrings);

            // array of doubles is okay
            double[] myDoubles = { 7.7 };

            PrintEnumerableValues<double[], double>(myDoubles);

            // custom type with indexer won't work with PrintEnumerableValues
            // because it doesn't implement the IEnumerable<T> constraint
            MyType<decimal> myType = new MyType<decimal>();
            myType.Add(9.9m);

            // doesn't compile because MyType is not IEnumerable<decimal>
            // which is the constraint on the PrintEnumerableValues method.
            //PrintEnumerableValues<MyType<decimal>, decimal>(myType);
        }

        /// <summary>
        /// Prints only elements of a collection
        /// that implements IEnumerable<T>"/>
        /// </summary>
        /// <typeparam name="T">Constrained as IEnumerable<U></typeparam>
        /// <typeparam name="U">Collection type to iterate through</typeparam>
        /// <param name="myItems">Collection of type T</param>
        private void PrintEnumerableValues<T, U>(T myItems)
            where T : IEnumerable<U>
        {
            foreach (U item in myItems)
            {
                Console.WriteLine(item);
            }
        }

        /// <summary>
        /// Algorithm to demonstrate and test
        /// interface constraints.
        /// </summary>
        public void TestBaseClassConstraints()
        {
            byte[] bytes = Encoding.Unicode.GetBytes(
                "The quick brown fox jumped over the lazy dog's back.");

            MemoryStream memStream = new MemoryStream(bytes);

            // MemoryStream is derived from Stream, which is okay
            PrintStreamValues<MemoryStream>(memStream);

            // StringReader is not a Stream, which will not work
            StringReader strRdr = new StringReader("I'm from a StringReader!");

            // compile-time error
            //PrintStreamValues<StringReader>(strRdr);
        }

        /// <summary>
        /// This will work with any Stream derived type
        /// </summary>
        /// <typeparam name="T">Must be a stream</typeparam>
        /// <param name="stream">To be read from</param>
        private void PrintStreamValues<T>(T stream)
            where T : Stream
        {
            int ByteSize = 1024;
            int count = 0;

            StringBuilder result = new StringBuilder();

            byte[] bytes = new byte[ByteSize];

            do
            {
                count = stream.Read(bytes, 0, ByteSize);

                // convert to string regardless of format
                result.Append(
                    Encoding.Unicode.GetString(bytes).
                    Substring(0, count - 1));
            }
            while (count == ByteSize);

            Console.WriteLine(result.ToString());
        }

        /// <summary>
        /// Exercise reference type and value type constraints
        /// </summary>
        public void TestReferenceAndValueTypeConstraints()
        {
            MyRefType refType = new MyRefType();
            refType.Val = 3;

            // modifies local reference
            CheckForReferenceType<MyRefType>(refType);

            Console.WriteLine(
                "refType.Val was 3 and is now " + refType.Val);

            MyValType valType = new MyValType();
            valType.Val = 3;

            // won't change local value
            CheckForValueType<MyValType>(valType);

            Console.WriteLine(
                "valType.Val was 3 and is now " + valType.Val);

            // compile-time errors

            // violates reference type constraint
            //			CheckForReferenceType<MyValType>(valType);

            // violates value type constraint
            //			CheckForValueType<MyRefType>(refType);
        }

        /// <summary>
        /// Used here as a constraint on methods that
        /// demonstrate constraints with reference
        /// type and value type constraints
        /// </summary>
        private interface IHoldVal
        {
            int Val
            {
                get;
                set;
            }
        }

        /// <summary>
        /// Used as an example class for working
        /// with reference type constraints and
        /// constructor constraint.
        /// </summary>
        public class MyRefType : IHoldVal
        {
            // must have default constructor
            public MyRefType() { }

            public MyRefType(int myParam)
            {
                val = myParam;
            }

            private int val;

            // implements IHoldVal
            public int Val
            {
                get { return val; }
                set { val = value; }
            }
        }

        /// <summary>
        /// Used as an example class for
        /// value type constraint demo and
        /// the constructor constraint demo.
        /// </summary>
        public struct MyValType : IHoldVal
        {
            public MyValType(int myParam)
            {
                val = myParam;
            }

            private int val;

            // implements IHoldVal
            public int Val
            {
                get { return val; }
                set { val = value; }
            }
        }


        /// <summary>
        /// Shows how to use a reference type constraint.
        /// </summary>
        /// <typeparam name="T">Type Parameter.</typeparam>
        /// <param name="type">Input Parameter.</param>
        private void CheckForReferenceType<T>(T type)
            where T : class, IHoldVal
        {
            type.Val = 5;
        }

        /// <summary>
        /// Shows how to use a value type constraint.
        /// </summary>
        /// <typeparam name="T">Type Parameter.</typeparam>
        /// <param name="type">Input Parameter.</param>
        private void CheckForValueType<T>(T type)
            where T : struct, IHoldVal
        {
            type.Val = 5;
        }

        /// <summary>
        /// Shows how constructor constraints work.
        /// </summary>
        public void TestConstructorConstraints()
        {
            // works with reference types
            IHoldVal myRef = CreateNewInstance<MyRefType>();
            Console.WriteLine(myRef.Val);

            // works with value types
            IHoldVal myVal = CreateNewInstance<MyValType>();
            Console.WriteLine(myVal.Val);
        }

        /// <summary>
        /// Creates a new object instance.
        /// </summary>
        /// <typeparam name="T">Type to create instance of.</typeparam>
        /// <returns>New instance.</returns>
        private T CreateNewInstance<T>()
            where T : IHoldVal, new()
        {
            T myNewType = new T();
            myNewType.Val = 7;

            // constructor constraint is for
            // default constructors only -
            // causes a compile-time error
            //			T anotherNewType = new T(7);

            return myNewType;
        }

        //private T CreateNewInstance<T>()
        //where T : IDisposable, new()
        //{
        //    T myNewType = new T();
        //    //myNewType.Val = 7;

        //    // constructor constraint is for
        //    // default constructors only -
        //    // causes a compile-time error
        //    //			T anotherNewType = new T(7);

        //    return myNewType;
        //}

    }

}